﻿package meyermike
{
    import flash.display.DisplayObjectContainer;
    import flash.utils.Dictionary;
    import flash.utils.getQualifiedClassName;

    /**
     * ...
     * @author Mike Meyer
     */
    public class GameObjectHolder
    {
        protected var mContainer:DisplayObjectContainer;
        protected var mLiveObjects:Array;

        protected var mReusables:Dictionary;

        public function GameObjectHolder(container:DisplayObjectContainer)
        {
            mContainer = container;
            mLiveObjects = new Array();
            mReusables = new Dictionary();
        }

        /**
         * add a game object to the holder.  unlike Add, it expects you to
         * provide a new instance and does not reuse objects
         */
        public function AddPremade(go:GameObject):void
        {
            mContainer.addChild(go);
            mLiveObjects.push(go);
        }


        /// Add a game object to the holder.  Will try to re-use existing ones
        public function Add(type:Class):GameObject
        {
            var go:GameObject = FindReusableObject(type);
            if (null == go)
            {
                go = new type();
            }

            AddPremade(go);
            go.Init();
            return go;
        }

        private function FindReusableObject(type:Class):GameObject
        {
            var classname:String = getQualifiedClassName(type);
            if (mReusables.hasOwnProperty(classname))
            {
                if (0 < mReusables[classname].length)
                {
                    return mReusables[classname].pop();
                }
            }

            return null;
        }

        public function Update(elapsedTime:Number):void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.Update(elapsedTime);
            }

            RemoveDeadObjects();
        }

        public function Integrate():void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.Integrate(Time.TimeDelta);
            }
        }

        /**
         * call after all the final positions have been figured
         * just before drawing.
         * Calls PreDraw on each layer.
         */
        public function PreDraw():void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.PreDraw();
            }
        }

        public function Beat(beat:Number):void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.Beat(beat);
            }
        }

        public function OffBeat(beat:Number):void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.OffBeat(beat);
            }
        }
        public function QuarterBeat(quarterBeat:Number):void
        {
            for each (var go:GameObject in mLiveObjects )
            {
                go.QuarterBeat(quarterBeat);
            }
        }

        private function RemoveDeadObjects():void
        {
            var removalQueue:Array = new Array();

            // look for any particles that need to be removed
            for (var i0:Number = 0; i0 < mLiveObjects.length; ++i0 )
            {
                var p0:GameObject = mLiveObjects[i0];

                // if it dies. queue it up for removal
                if (!p0.IsAlive())
                {
                    removalQueue.push(i0);
                }
            }

            // do any queued removals. looping backwards! important!
            for (var i1:Number = (removalQueue.length - 1); i1 >= 0 ; --i1)
            {
                RemoveAt(removalQueue[i1]);
            }
        }

        /**
         * remove the game object at index from the scene.  Store it in
         * reusables for later reuse.
         */
        private function RemoveAt(index:Number):void
        {
            var p:GameObject = mLiveObjects[index];
            p.Shutdown();

            var classname:String = getQualifiedClassName(p);

            // never reuse the base gameobject class, that will probably
            // always have custom stuff attached in it.
            if ("meyermike::GameObject" != classname)
            {
                if (!mReusables.hasOwnProperty(classname))
                {
                    mReusables[classname] = new Array();
                }
                mReusables[classname].push(p);
            }

            mContainer.removeChild(p);
            mLiveObjects.splice(index, 1);
        }

        /**
         * call a function for each go in the layer.  func should be of the form
         * function func(go:GameObject):void
         */
        public function ForEach(func:Function):void
        {
            for (var i:Number = 0; i < mLiveObjects.length; ++i )
            {
                func(mLiveObjects[i]);
            }
        }

        /**
         * Get rid of any and all objects held by this thing
         */
        public function RemoveAll():void
        {
            for (var i:Number = (mLiveObjects.length - 1); i >= 0 ; --i)
            {
                RemoveAt(i);
            }
        }

        /**
         * Return the number of LIVE objects (not cached, but currently unused ones)
         */
        public function GetNumObjects():Number
        {
            return this.mLiveObjects.length;
        }

        /**
         * get a (live) object by it's index
         */
        public function GetObject(index:Number):GameObject
        {
            return mLiveObjects[index];
        }

        /**
         * access to the container movieclip
         */
        public function GetContainer():DisplayObjectContainer
        {
            return this.mContainer;
        }
    }

}
